#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "midi_b9.h"

/* **************************************************************** */

int detecmap(float x_source, float y_source, char srcfun[], int last, int jemx, 
   	float *normv, struct euclid *g_const, struct instr_data *jmx_id,
        struct log_con *logger, struct backpro *backproj)
 
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                                               */        
/*      detecmap.c generates a list of detector pixels which are                 */        
/*      illuminated from a particular sky position.                              */
/*                                                                               */
/*      Niels Lund, February 2003                                                */        
/*      Updated for NJW-mask and detector support struct. Sep. 2003              */
/*                                                                               */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* Illumination JEM-X detector surface for source at position (x_source, y_source). 
   Derived from skymap version 3.0 
	
*	Niels Lund, January 2003	Version 1.0
*	Danish Space Research Institute
*	
*	NOTE: The jemx variable is 0 for JEM-X1 and 1 for JEM-X2 in these routines !!
*
*	The XY-coordinate system used in this routine is aligned with 
*	the detector collimator lamellae. The microstrip pattern is 
*	approximately aligned with the lamellae. The cathode positions 
*	are (nearly) along the X-axis, the backside positions nearly 
*	along the Y-axis. The center of the detector is located at
*	position (128, 128). The central collimator crossing is
*	also at this position. The  center of the collimator falls at 
*       the corner between 4 pixel boundaries.
*
*	The unit used for the X- and Y-positions on the detector are 
*	1 mm. Actually, the X- and Y-integer positions are only 
*	approximately separated by 1 mm. A detailed map, specifying 
*   	the float values of the X- and Y-position of the four corners 
*	of each pixel have been prepared by Carl Budtz-J and are
*	applied to the data using subroutine "pixelfold". 
*
*	The full field of view of the JEM-X detector is defined by the
*	dimensions of the mask (535 mm diameter) and detector (250 mm
*	diameter) and by the mask-detector separation (about 3400 mm).
*
*       The full field of view, which has a radius of 'FOV_radius' pixels. 
*
*/

{
   int i, j, k, m, allownum, n, ring, nn, dif, sky2, iq, iv, ky, mx, mmm, iqp, ivp;
   int x20, y20, xy20, looplim, xcdif, ycdif, xyacc, p_loss, allowpix=0, njw1, njw2;
   int njwhole, strblock, njwOK, p_count=0;

   unsigned det_index, dd_index, sky_index, ss_index, r_sky2;
   unsigned mx2,  xd7, yd7, xc7, yc7, cur_index, aind;
   unsigned st_byt_in;

   static double collp30;	/* 30 times the collimator pitch */
   static double cosal, sinal, cosbe, sinbe, cosga, singa;
   static double rsin60, cot60;
   static double coll_ptop, collp30top;

   double ic, jc, coll_r2, xdet, ydet, xmask, ymask;
   double allow, r_sky, c_fact, c_factb, m_fact, mask_r2, rm2;
   double ydet2, qmask, vmask, dif7, xcoll2;
   double xdmod7, ydmod7, xtmod7, ytmod7, c0_fact, sum, xtemp;

   double ycolbot, ycoltop, xcolbot, xcoltop, xrmask, yrmask, xmask2;
   double xrcbot, xrctop, yrcbot, yrctop, xrdet, yrdet;
   float norm;
   static int func_local = -1, local_calling_num=0;
   clock_t clock(void), calling_TT, TT_now;

   /* function entry timing start */
   if (func_local < 0) {
      func_local = logger->func_used_num++;
      sprintf(logger->func_names[func_local], "detecmap");
      if (logger->func_used_num > 199) {printf("func_used_num > 199\n"); exit(0);}
   }
   logger->func_calls[func_local]++;
   calling_TT = logger->TT;
   logger->TT = TT_now = clock();
   local_calling_num = logger->func_calling_num;
   logger->func_calling_num = func_local;
   logger->func_times[local_calling_num] += difftime(TT_now, calling_TT);
   /* function entry timing end */
   
   
   /* initializations */

   collp30 = jmx_id->coll_pitch * 30.0;		/* 30 time the collimator pitch, used in the "modulo" calc */

   coll_ptop = jmx_id->coll_pitch * jmx_id->coll_curv;
   collp30top = coll_ptop * 30.0;

   coll_r2 = jmx_id->coll_r * jmx_id->coll_r;
   mask_r2 = jmx_id->mask_r * jmx_id->mask_r;

   cosal = cos(jmx_id->alfa);
   sinal = sin(jmx_id->alfa);
   cosbe = cos(jmx_id->beta);
   sinbe = sin(jmx_id->beta);
   cosga = cos(jmx_id->gammav);
   singa = sin(jmx_id->gammav);
   
   r_sky = jmx_id->mask_r + jmx_id->coll_r;	/* field of view radius */

/*   c0_fact = (jmx_id->coll_pitch - 2.0 * jmx_id->coll_w) / backproj->FOV_radius;   */
   c0_fact = 6.6 / backproj->FOV_radius;	/* stepsize in (mm) in skymap at collimator top level */ 
      							/* (0.0519685 mm for 'FOV_radius' = 127) */
                 /* this routine calculates illumination at the detector window level (indepc = 0.0) */
   m_fact = c0_fact * jmx_id->mask_height  / jmx_id->coll_h;	/* corresponding stepsize at mask level */	
   c_factb = 0.0;	/* corresponding stepsize at bottom of collimator 
      				(this calculation is valid for the level of the detector window) */
   c_fact = c0_fact;					/* corresponding stepsize at top of collimator */
   
   if (logger->test == 1) {
      sprintf(logger->tmptxt, 
         "detecmap: c_fact: %6.2lf, m_fact: %6.2lf, c_factb: %6.2lf, x/y_source: %6.1f/%6.1f\n",
         c_fact, m_fact, c_factb, x_source, y_source);
      logger->logstat = logprint(logger->tmptxt, logger->lognum);
   }
   cot60 = 1.0 / (jmx_id->hexa_dim * 1.73205081);
   rsin60 = 2.0 * cot60;

   det_index = 0;
   sky_index = 0;
   cur_index = 0;
   allownum = 0;

    /* ***************  main loop ********************* */

   x_source -= (float)(backproj->FOV_radius);
   y_source -= (float)(backproj->FOV_radius);
         
   for (i=0; i<256; i++) {		/* loop over detector Y */

      yrdet = (double)(i) - 127.5 + jmx_id->yshi;
      ydet2 = yrdet * yrdet;
      
      for (j=0; j<256; j++) {		/* loop over detector X */
         
	 det_index = j * 256 + i;
         srcfun[det_index] = 0;

	 xrdet = (double)(j) - 127.5 + jmx_id->xshi;
         if ((ydet2 + xrdet*xrdet) > coll_r2) continue; 		/* detector radius check */
         
         xdet = xrdet * cosga - yrdet * singa;
         ydet = yrdet * cosga + xrdet * singa;
         
	 /* xdet and ydet are coordinates of a detector pixel in a coord. system aligned after Common-X */

         xyacc = 0;
         
         mx = x_source;
         mx2 = mx * mx;

         xcolbot = x_source * c_factb + xdet;
         xcoltop = x_source * c_fact + xdet;
         xcoll2 = xcoltop * xcoltop;
            
         xrmask = x_source * m_fact + xdet;
         xmask2 = xrmask * xrmask;
                
         ky = y_source;
                
         ycolbot = y_source * c_factb + ydet;
         ycoltop = y_source * c_fact + ydet;
            
         if ((xcoll2 + ycoltop * ycoltop) > coll_r2) continue;
         				/* Collimator top radius check */
                
         yrmask = y_source * m_fact + ydet;
            
         rm2 = yrmask * yrmask + xmask2;
                
         if (rm2 > mask_r2) continue;			/* the event falls outside the coded mask */
         
         srcfun[det_index] = 1;		/* indicate area on which the mask pattern is projected */

         xrcbot = xcolbot * cosbe + ycolbot * sinbe;	/* rotation through -jmx_id->beta! */
         yrcbot = ycolbot * cosbe - xcolbot * sinbe;
            
         xrctop = xcoltop * cosbe + ycoltop * sinbe;	/* rotation through -jmx_id->beta! */
         yrctop = ycoltop * cosbe - xcoltop * sinbe;
            
         /* determine open directions of collimator the collimator lamellae are located on positions:
            -126.9,-119.85, ... ,-21.15, -14.1, -7.05, 0, 7.05, 14.1, 21.15, ... , 119.85, 126.9
            both in the X- and Y-direction 
         */

         yd7 = (int)((yrcbot + collp30) / jmx_id->coll_pitch);	/* determine bottom collimator hole index
                   							   collp30 >detector radius */
         yc7 = (int)((yrctop + jmx_id->c_tilty + collp30top) / coll_ptop); /* determine top collimator hole index */
	    
         
         if (yc7 > yd7) continue;
               /* the event is lost due to wall crossing in the collimator */
                   
         if (yc7 < yd7) continue;
               /* the event is lost due to wall crossing in the collimator */
                   
         ydmod7 = yrcbot + collp30 - jmx_id->coll_pitch * yd7;	/* get remainder Ycoord within this hole */

         if (ydmod7 < jmx_id->coll_wb) continue;
               /* the event is lost at the bottom edge of the collimator */
                   	   
         if (ydmod7 > (jmx_id->coll_pitch - jmx_id->coll_wb)) continue;
               /* the event is lost at the bottom edge of the collimator */

         ytmod7 = yrctop + jmx_id->c_tilty + collp30top - coll_ptop * yc7;	/* remainder Ycoord within this hole */
               
         if (ytmod7 < jmx_id->coll_w) continue;
               /* the event is lost at the top edge of the collimator */
					/* Y-collimator wall -thickness check */
                   	   
         if (ytmod7 > (coll_ptop - jmx_id->coll_w)) continue;
               /* the event is lost at the top edge of the collimator */
					/* Y-collimator wall +thickness check */         	   

         xd7 = (int)((xrcbot + collp30) / jmx_id->coll_pitch);
         xc7 = (int)((xrctop + jmx_id->c_tiltx + collp30top) / coll_ptop);
                    
         if (xc7 > xd7) continue;
               /* the event is lost due to wall crossing in the collimator */
                    
         if (xc7 < xd7) continue;
               /* the event is lost due to wall crossing in the collimator */

         xdmod7 = xrcbot + collp30 - jmx_id->coll_pitch * xd7;
            
         if (xdmod7 < jmx_id->coll_wb) continue;
               /* the event is lost at the bottom edge of the collimator */

         if (xdmod7 > (jmx_id->coll_pitch - jmx_id->coll_wb)) continue;
               /* the event is lost at the bottom edge of the collimator */

         xtmod7 = xrctop + jmx_id->c_tiltx + collp30top - coll_ptop * xc7;
            
         if (xtmod7 < jmx_id->coll_w) continue;
               /* the event is lost at the top edge of the collimator */

         if (xtmod7 > (coll_ptop - jmx_id->coll_w)) continue;
               /* the event is lost at the top edge of the collimator */

         xmask = xrmask * cosal + yrmask * sinal; 
         ymask = yrmask * cosal - xrmask * sinal;

         qmask = xmask / jmx_id->hexa_dim - ymask * cot60;
         vmask = ymask * rsin60;

   	 iqp = qmask + 22501.5;
         ivp = vmask + 22501.5;
         iq = iqp - 22501;
         iv = ivp - 22501;

         njw1 = 99 + iq + iv;
         njw2 = 100 + iv;
         
         if ((njw1 < 0) || (njw1 > 199) || (njw1 < 0) || (njw1 > 199)) {
            if (p_count < 3) {
               sprintf(logger->tmptxt, 
                  "njw1/2 outside limits: %4d/%4d, i/j: %3d/%3d, x/y_source: %6.2f/%6.2f\n",
                  njw1, njw2, i, j, x_source, y_source);
               logger->logstat = logprint(logger->tmptxt, logger->lognum);
               p_count++;
            }
            njw1 = njw2 = 0;   
         }
         
         if (jmx_id->njwmask[njw1][njw2] == 6) {
            njwOK++;
            logger->lognum = 2;
            strblock = struchk(xmask, ymask, xdet, ydet, last, g_const, jmx_id, logger);
            if (strblock < 0) return(-1);
            logger->lognum = 0;
            if (strblock == 0) { 
               srcfun[det_index] += 1;
               allownum++;   
            }

         }
      }
   }
   if (last) struchk(30.0, 30.0, 1.0, 1.0, 2, g_const, jmx_id, logger);
   /* dummy call to initiate statistics  print */
   
   if (logger->test == 1) {
      sprintf(logger->tmptxt, "detecmap complete: allownum: %6d, srcfun: ", allownum);
      sprintf(logger->tmptxt2, "%2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\n",
         srcfun[detdim/2 - 6], srcfun[detdim/2 - 5], srcfun[detdim/2 - 4], srcfun[detdim/2 - 3], 
         srcfun[detdim/2 - 2], srcfun[detdim/2 - 1], srcfun[detdim/2 - 0], srcfun[detdim/2 + 1], 
         srcfun[detdim/2 + 2], srcfun[detdim/2 +3], srcfun[detdim/2 + 4], srcfun[detdim/2 + 5]);
      strcat(logger->tmptxt, logger->tmptxt2);
      logger->logstat = logprint(logger->tmptxt, logger->lognum);
   }
   if ( allownum > 0) *normv = 1.0 / (float)(allownum);
   else *normv = -1.0;

   /* function exit timing start */
   TT_now = clock();
   logger->func_times[func_local] += difftime(TT_now, logger->TT);
   logger->TT = TT_now;
   logger->func_calling_num = local_calling_num;
   /* function exit timing end */
   
   return(0);
}

